import datetime
import json
import logging
import os
import tempfile
import time
from collections import UserDict, UserList
from copy import deepcopy
from typing import Any, List, Union

import jsonpickle

from .config import config, s3_config
from .io import S3Artifact
from .op_template import get_k8s_client
from .utils import download_artifact, get_key, upload_s3

try:
    import kubernetes
except Exception:
    pass

logger = logging.getLogger(__name__)


class ArgoObjectDict(UserDict):
    """
    Generate ArgoObjectDict and ArgoObjectList on initialization rather than
    on __getattr__, otherwise modify a.b.c will not take effect
    """

    def __init__(self, d):
        super().__init__(d)
        for key, value in self.items():
            if isinstance(value, dict):
                self.data[key] = ArgoObjectDict(value)
            elif isinstance(value, list):
                self.data[key] = ArgoObjectList(value)

    def __getattr__(self, key):
        if key == "data":
            return super().__getattr__(key)

        if key in self.data:
            return self.data[key]
        else:
            raise AttributeError(
                "'ArgoObjectDict' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        if key == "data":
            return super().__setattr__(key, value)

        self.data[key] = value

    def recover(self):
        return {key: value.recover() if isinstance(value, (ArgoObjectDict,
                                                           ArgoObjectList))
                else value for key, value in self.data.items()}


class ArgoObjectList(UserList):
    def __init__(self, li):
        super().__init__(li)
        for i, value in enumerate(self.data):
            if isinstance(value, dict):
                self.data[i] = ArgoObjectDict(value)
            elif isinstance(value, list):
                self.data[i] = ArgoObjectList(value)

    def recover(self):
        return [value.recover() if isinstance(value, (ArgoObjectDict,
                                                      ArgoObjectList))
                else value for value in self.data]


class ArgoParameter(ArgoObjectDict):
    def __init__(self, par):
        super().__init__(par)

    def __getattr__(self, key):
        if ((key == "value" and "value" not in self.data) or
            (key == "type" and "type" not in self.data)) and \
                hasattr(self, "save_as_artifact"):
            with tempfile.TemporaryDirectory() as tmpdir:
                try:
                    download_artifact(self, path=tmpdir)
                    fs = os.listdir(tmpdir)
                    assert len(fs) == 1
                    with open(os.path.join(tmpdir, fs[0]), "r") as f:
                        content = jsonpickle.loads(f.read())
                        self.value = content
                except Exception as e:
                    logger.warning("Failed to load parameter value from "
                                   "artifact: %s" % e)
        if key == "value" and hasattr(self, "description") and \
                self.description is not None:
            desc = jsonpickle.loads(self.description)
            if desc["type"] != str(str):
                try:
                    return jsonpickle.loads(super().__getattr__("value"))
                except Exception as e:
                    logger.warning("Failed to unpickle parameter: %s" % e)
        return super().__getattr__(key)


class ArgoStep(ArgoObjectDict):
    def __init__(self, step, workflow):
        super().__init__(deepcopy(step))
        self.workflow = workflow
        self.pod = None
        self.key = None
        if hasattr(self, "inputs"):
            self.handle_io(self.inputs)
            if hasattr(self.inputs, "parameters") and "dflow_key" in \
                    self.inputs.parameters and self.inputs.parameters[
                        "dflow_key"].value != "":
                self.key = self.inputs.parameters["dflow_key"].value

        if hasattr(self, "outputs"):
            self.handle_io(self.outputs)

    def handle_io(self, io):
        if hasattr(io, "parameters") and \
                isinstance(io.parameters, ArgoObjectList):
            io.parameters = {par.name: ArgoParameter(par)
                             for par in io.parameters}

        if hasattr(io, "artifacts") and \
                isinstance(io.artifacts, ArgoObjectList):
            io.artifacts = {art.name: art for art in io.artifacts}

        self.handle_big_parameters(io)

    def handle_big_parameters(self, io):
        if hasattr(io, "artifacts"):
            for name, art in io.artifacts.items():
                if name[:13] == "dflow_bigpar_":
                    if not hasattr(io, "parameters"):
                        io.parameters = {}
                    if name[13:] not in io.parameters:
                        par = art.copy()
                        par["name"] = name[13:]
                        par["save_as_artifact"] = True
                        io.parameters[name[13:]] = ArgoParameter(par)

    def modify_output_parameter(
            self,
            name: str,
            value: Any,
    ) -> None:
        """
        Modify output parameter of an Argo step

        Args:
            name: parameter name
            value: new value
        """
        if isinstance(value, str):
            self.outputs.parameters[name].value = value
        else:
            self.outputs.parameters[name].value = jsonpickle.dumps(value)

        if hasattr(self.outputs.parameters[name], "save_as_artifact"):
            with tempfile.TemporaryDirectory() as tmpdir:
                path = tmpdir + "/" + name
                with open(path, "w") as f:
                    f.write(jsonpickle.dumps(value))
                key = upload_s3(path)
                s3 = S3Artifact(key=key)
                if s3_config["repo_type"] == "s3":
                    self.outputs.artifacts["dflow_bigpar_" + name].s3 = \
                        ArgoObjectDict(s3.to_dict())
                elif s3_config["repo_type"] == "oss":
                    self.outputs.artifacts["dflow_bigpar_" + name].oss = \
                        ArgoObjectDict(s3.oss().to_dict())

    def modify_output_artifact(
            self,
            name: str,
            s3: S3Artifact,
    ) -> None:
        """
        Modify output artifact of an Argo step

        Args:
            name: artifact name
            s3: replace the artifact with a s3 object
        """
        if config["mode"] == "debug":
            self.outputs.artifacts[name].local_path = s3.local_path
            return
        assert isinstance(s3, S3Artifact), "must provide a S3Artifact object"
        old_key = get_key(self.outputs.artifacts[name], raise_error=False)
        if s3_config["repo_type"] == "s3":
            self.outputs.artifacts[name].s3 = ArgoObjectDict(s3.to_dict())
        elif s3_config["repo_type"] == "oss":
            self.outputs.artifacts[name].oss = ArgoObjectDict(
                s3.oss().to_dict())
        if s3.key[-4:] == ".tgz" and hasattr(self.outputs.artifacts[name],
                                             "archive"):
            del self.outputs.artifacts[name]["archive"]
        elif s3.key[-4:] != ".tgz" and not hasattr(self.outputs.artifacts[
                name], "archive"):
            self.outputs.artifacts[name]["archive"] = {"none": {}}
        self.outputs.artifacts[name].modified = {"old_key": old_key}

    def retry(self):
        from .workflow import Workflow, get_argo_api_client
        wf = Workflow(id=self.workflow)
        assert wf.query_status() == "Running"
        logger.info("Suspend workflow %s..." % self.workflow)
        wf.suspend()
        time.sleep(5)

        logger.info("Query workflow %s..." % self.workflow)
        wf_info = wf.query().recover()
        nodes = wf_info["status"]["nodes"]
        nodes[self.id]["phase"] = "Pending"
        for node in nodes.values():
            if self.name.startswith(node["name"]) and \
                    node["phase"] == "Failed":
                node["phase"] = "Running"

        logger.info("Delete pod of step %s..." % self.id)
        self.delete_pod()
        with get_argo_api_client() as api_client:
            logger.info("Update workflow %s..." % self.workflow)
            api_client.call_api(
                '/api/v1/workflows/%s/%s' % (
                    config["namespace"], self.workflow),
                'PUT', response_type='object',
                header_params=config["http_headers"],
                body={"workflow": wf_info},
                _return_http_data_only=True)

        logger.info("Resume workflow %s..." % self.workflow)
        wf.resume()

    def get_pod(self):
        assert self.type == "Pod"
        wf_name = self.workflow
        node_name = self.name
        template_name = self.templateName
        node_id = self.id
        pod_name = get_pod_name(wf_name, node_name, template_name, node_id)
        with get_k8s_client() as k8s_client:
            core_v1_api = kubernetes.client.CoreV1Api(k8s_client)
            self.pod = core_v1_api.api_client.call_api(
                '/api/v1/namespaces/%s/pods/%s' % (config["namespace"],
                                                   pod_name),
                'GET', response_type='V1Pod',
                header_params=config["http_headers"],
                _return_http_data_only=True)
        return self.pod

    def delete_pod(self):
        assert self.type == "Pod"
        wf_name = self.workflow
        node_name = self.name
        template_name = self.templateName
        node_id = self.id
        pod_name = get_pod_name(wf_name, node_name, template_name, node_id)
        with get_k8s_client() as k8s_client:
            core_v1_api = kubernetes.client.CoreV1Api(k8s_client)
            try:
                core_v1_api.api_client.call_api(
                    '/api/v1/namespaces/%s/pods/%s' % (config["namespace"],
                                                       pod_name),
                    'DELETE', response_type='V1Pod',
                    header_params=config["http_headers"],
                    _return_http_data_only=True)
                logger.info("Deleted pod %s" % pod_name)
            except Exception as e:
                logging.warning("Failed to delete pod %s: %s" % (pod_name, e))
            try:
                while True:
                    core_v1_api.api_client.call_api(
                        '/api/v1/namespaces/%s/pods/%s' % (
                            config["namespace"], pod_name),
                        'GET', response_type='V1Pod',
                        header_params=config["http_headers"],
                        _return_http_data_only=True)
                    logger.info("Waiting pod %s to be deleted..." % pod_name)
                    time.sleep(1)
            except Exception:
                pass

    def get_script(self):
        if self.pod is None:
            self.get_pod()
        main_container = next(filter(lambda c: c.name == "main",
                                     self.pod.spec.containers))
        templ_env = next(filter(lambda e: e.name == "ARGO_TEMPLATE",
                                main_container.env))
        templ = json.loads(templ_env.value)
        return templ["script"]["source"]

    def set_script(self, script):
        if self.pod is None:
            self.get_pod()
        main_container = next(filter(lambda c: c.name == "main",
                                     self.pod.spec.containers))
        templ_env = next(filter(lambda e: e.name == "ARGO_TEMPLATE",
                                main_container.env))
        templ = json.loads(templ_env.value)
        templ["script"]["source"] = script
        templ_env.value = json.dumps(templ)
        init_container = next(filter(lambda c: c.name == "init",
                                     self.pod.spec.init_containers))
        templ_env = next(filter(lambda e: e.name == "ARGO_TEMPLATE",
                                init_container.env))
        templ_env.value = json.dumps(templ)
        wait_container = next(filter(lambda c: c.name == "wait",
                                     self.pod.spec.containers))
        templ_env = next(filter(lambda e: e.name == "ARGO_TEMPLATE",
                                wait_container.env))
        templ_env.value = json.dumps(templ)

    def replay(self):
        if self.pod is None:
            self.get_pod()
        self.pod.metadata.resource_version = None
        self.pod.spec.node_name = None
        self.delete_pod()
        with get_k8s_client() as k8s_client:
            core_v1_api = kubernetes.client.CoreV1Api(k8s_client)
            core_v1_api.api_client.call_api(
                '/api/v1/namespaces/%s/pods' % config["namespace"],
                'POST', body=self.pod, response_type='V1Pod',
                header_params=config["http_headers"],
                _return_http_data_only=True)

    def get_duration(self) -> datetime.timedelta:
        return get_duration(self)


max_k8s_resource_name_length = 253
k8s_naming_hash_length = 10


def get_pod_name(wf_name, node_name, template_name, node_id):
    if wf_name == node_name:
        return wf_name
    prefix = "%s-%s" % (wf_name, template_name)
    max_prefix_length = max_k8s_resource_name_length - k8s_naming_hash_length
    if len(prefix) > max_prefix_length - 1:
        prefix = prefix[:max_prefix_length-1]
    hash_val = fnva(node_name.encode(), FNV1_32_INIT, FNV_32_PRIME, 2**32)
    return "%s-%s" % (prefix, hash_val)


FNV_32_PRIME = 0x01000193
FNV1_32_INIT = 0x811c9dc5


def fnva(data, hval_init, fnv_prime, fnv_size):
    """
    Alternative FNV hash algorithm used in FNV-1a.
    """
    assert isinstance(data, bytes)

    hval = hval_init
    for byte in data:
        hval = hval ^ byte
        hval = (hval * fnv_prime) % fnv_size
    return hval


class ArgoWorkflow(ArgoObjectDict):
    def __init__(self, d):
        super().__init__(d)
        self.id = None
        self.uid = None
        if hasattr(self, "metadata") and hasattr(self.metadata, "name"):
            self.id = self.metadata.name
        if hasattr(self, "metadata") and hasattr(self.metadata, "uid"):
            self.uid = self.metadata.uid

    def get_step(
            self,
            name: Union[str, List[str]] = None,
            key: Union[str, List[str]] = None,
            phase: Union[str, List[str]] = None,
            id: Union[str, List[str]] = None,
            type: Union[str, List[str]] = None,
    ) -> List[ArgoStep]:
        if name is not None and not isinstance(name, list):
            name = [name]
        if key is not None and not isinstance(key, list):
            key = [key]
        if phase is not None and not isinstance(phase, list):
            phase = [phase]
        if id is not None and not isinstance(id, list):
            id = [id]
        if type is not None and not isinstance(type, list):
            type = [type]
        step_list = []
        if hasattr(self.status, "nodes"):
            for step in self.status.nodes.values():
                if step["startedAt"] is None:
                    continue
                if name is not None and not match(step["displayName"], name):
                    continue
                if key is not None:
                    step_key = None
                    if "inputs" in step and "parameters" in step["inputs"]:
                        for par in step["inputs"]["parameters"]:
                            if par["name"] == "dflow_key":
                                step_key = par["value"]
                    if step_key not in key:
                        continue
                if phase is not None and not ("phase" in step and
                                              step["phase"] in phase):
                    continue
                if type is not None and not ("type" in step and
                                             step["type"] in type):
                    continue
                if id is not None and step["id"] not in id:
                    continue
                step = ArgoStep(step, self.metadata.name)
                step_list.append(step)
        step_list.sort(key=lambda x: x["startedAt"])
        return step_list

    def get_duration(self) -> datetime.timedelta:
        return get_duration(self.status)


def get_duration(status) -> datetime.timedelta:
    if status.startedAt is None:
        return datetime.timedelta()
    else:
        ts = datetime.datetime.strptime(status.startedAt,
                                        "%Y-%m-%dT%H:%M:%SZ")
        if status.finishedAt is None:
            tf = datetime.datetime.now()
        else:
            tf = datetime.datetime.strptime(status.finishedAt,
                                            "%Y-%m-%dT%H:%M:%SZ")
        return tf - ts


def match(n, names):
    for name in names:
        if n == name or n.find(name + "(") == 0:
            return True
    return False



import argparse
import datetime
from typing import List, Optional

from dflow import (S3Artifact, Secret, Workflow, config, download_artifact,
                   query_workflows, upload_artifact)


def main_parser():
    parser = argparse.ArgumentParser(
        description="dflow: a Python framework for constructing scientific "
        "computing workflows",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        title="Valid subcommands", dest="command")

    parser_list = subparsers.add_parser(
        "list",
        help="List workflows",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_list.add_argument(
        "-l",
        "--label",
        type=str,
        default=None,
        help="query by labels",
    )

    parser_get = subparsers.add_parser(
        "get",
        help="Get a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_get.add_argument("ID", help="the workflow ID.")

    parser_getsteps = subparsers.add_parser(
        "getsteps",
        help="Get steps from a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_getsteps.add_argument("ID", help="the workflow ID.")
    parser_getsteps.add_argument(
        "-n",
        "--name",
        type=str,
        default=None,
        help="query by name",
    )
    parser_getsteps.add_argument(
        "-k",
        "--key",
        type=str,
        default=None,
        help="query by key",
    )
    parser_getsteps.add_argument(
        "-p",
        "--phase",
        type=str,
        default=None,
        help="query by phase",
    )
    parser_getsteps.add_argument(
        "-i",
        "--id",
        type=str,
        default=None,
        help="query by ID",
    )
    parser_getsteps.add_argument(
        "-t",
        "--type",
        type=str,
        default=None,
        help="query by type",
    )

    parser_getkeys = subparsers.add_parser(
        "getkeys",
        help="Get keys of steps from a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_getkeys.add_argument("ID", help="the workflow ID.")

    parser_delete = subparsers.add_parser(
        "delete",
        help="Delete a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_delete.add_argument("ID", help="the workflow ID.")

    parser_resubmit = subparsers.add_parser(
        "resubmit",
        help="Resubmit a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_resubmit.add_argument("ID", help="the workflow ID.")

    parser_resume = subparsers.add_parser(
        "resume",
        help="Resume a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_resume.add_argument("ID", help="the workflow ID.")

    parser_retry = subparsers.add_parser(
        "retry",
        help="Retry a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_retry.add_argument("ID", help="the workflow ID.")
    parser_retry.add_argument(
        "-s",
        "--step",
        type=str,
        default=None,
        help="retry a step in a running workflow with step ID (experimental)",
    )

    parser_stop = subparsers.add_parser(
        "stop",
        help="Stop a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_stop.add_argument("ID", help="the workflow ID.")

    parser_suspend = subparsers.add_parser(
        "suspend",
        help="Suspend a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_suspend.add_argument("ID", help="the workflow ID.")

    parser_terminate = subparsers.add_parser(
        "terminate",
        help="Terminate a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_terminate.add_argument("ID", help="the workflow ID.")

    parser_wait = subparsers.add_parser(
        "wait",
        help="Wait a workflow to complete",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_wait.add_argument("ID", help="the workflow ID.")
    parser_wait.add_argument(
        "-i",
        "--interval",
        type=int,
        default=1,
        help="time interval between two queries",
    )

    parser_download = subparsers.add_parser(
        "download",
        help="Download an artifact",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_download.add_argument(
        "-k",
        "--key",
        type=str,
        default=None,
        help="storage key of the artifact",
    )
    parser_download.add_argument(
        "-u",
        "--urn",
        type=str,
        default=None,
        help="URN of the artifact",
    )
    parser_download.add_argument(
        "-p",
        "--path",
        type=str,
        default=".",
        help="the path to which the artifact will be downloaded",
    )

    parser_upload = subparsers.add_parser(
        "upload",
        help="Upload an artifact",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_upload.add_argument(
        "-p",
        "--path",
        type=str,
        required=True,
        help="path of the local file(s)",
    )
    parser_upload.add_argument(
        "-n",
        "--namespace",
        type=str,
        default=None,
        help="namespace for registering dataset",
    )
    parser_upload.add_argument(
        "-d",
        "--name",
        type=str,
        default=None,
        help="dataset name for registering dataset",
    )
    parser_upload.add_argument(
        "-s",
        "--description",
        type=str,
        default=None,
        help="description for registering dataset",
    )
    parser_upload.add_argument(
        "-t",
        "--tag",
        type=str,
        default=None,
        help="tags for registering dataset",
    )
    parser_upload.add_argument(
        "-r",
        "--property",
        type=str,
        default=None,
        help="properties for registering dataset",
    )

    parser_submit = subparsers.add_parser(
        "submit",
        help="Submit a workflow from a YAML file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_submit.add_argument("FILE", help="the YAML file")
    parser_submit.add_argument(
        "-d",
        "--detach",
        action="store_true",
        help="detach mode for running workflow",
    )
    parser_submit.add_argument(
        "-n",
        "--name",
        type=str,
        default=None,
        help="workflow ID",
    )

    parser_create = subparsers.add_parser(
        "create",
        help="Create a resource",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    create_subparsers = parser_create.add_subparsers(
        title="Valid resources", dest="resource")
    parser_secret = create_subparsers.add_parser(
        "secret",
        help="Create a secret",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_secret.add_argument("value", help="secret value")
    parser_secret.add_argument(
        "-n",
        "--name",
        type=str,
        help="name of the secret",
    )
    parser_secret.add_argument(
        "-k",
        "--key",
        type=str,
        default="secret",
        help="key in the secret",
    )
    return parser


def parse_args(args: Optional[List[str]] = None):
    """Commandline options argument parsing.

    Parameters
    ----------
    args : List[str]
        list of command line arguments, main purpose is testing default option
        None takes arguments from sys.argv
    """
    parser = main_parser()
    parsed_args = parser.parse_args(args=args)
    if parsed_args.command is None:
        parser.print_help()
    return parsed_args


def format_print_table(t: List[List[str]]):
    ncol = len(t[0])
    maxlen = [0] * ncol
    for row in t:
        for i, s in enumerate(row):
            if len(str(s)) > maxlen[i]:
                maxlen[i] = len(str(s))
    for row in t:
        for i, s in enumerate(row):
            print(str(s) + " " * (maxlen[i]-len(str(s))+3), end="")
        print()


def format_time_delta(td: datetime.timedelta) -> str:
    if td.days > 0:
        return "%dd" % td.days
    elif td.seconds >= 3600:
        return "%dh" % (td.seconds // 3600)
    else:
        return "%ds" % td.seconds


def main():
    args = parse_args()

    if args.command == "list":
        if args.label is not None:
            labels = {}
            for label in args.label.split(","):
                key, value = label.split("=")
                labels[key] = value
        else:
            labels = None
        wfs = query_workflows(labels=labels)
        t = [["NAME", "STATUS", "AGE", "DURATION"]]
        for wf in wfs:
            tc = datetime.datetime.strptime(wf.metadata.creationTimestamp,
                                            "%Y-%m-%dT%H:%M:%SZ")
            age = format_time_delta(datetime.datetime.now() - tc)
            dur = format_time_delta(wf.get_duration())
            t.append([wf.id, wf.status.phase, age, dur])
        format_print_table(t)
    elif args.command == "get":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        info = wf.query()
        t = []
        t.append(["Name:", info.id])
        t.append(["Status:", info.status.phase])
        t.append(["Created:", info.metadata.creationTimestamp])
        t.append(["Started:", info.status.startedAt])
        t.append(["Finished:", info.status.finishedAt])
        t.append(["Duration", format_time_delta(info.get_duration())])
        t.append(["Progress:", info.status.progress])
        format_print_table(t)
        print()
        steps = info.get_step()
        t = [["STEP", "ID", "KEY", "TYPE", "PHASE", "DURATION"]]
        for step in steps:
            if step.type in ["StepGroup"]:
                continue
            key = step.key if step.key is not None else ""
            dur = format_time_delta(step.get_duration())
            t.append([step.displayName, step.id, key, step.type, step.phase,
                      dur])
        format_print_table(t)
    elif args.command == "getsteps":
        wf_id = args.ID
        name = args.name
        key = args.key
        phase = args.phase
        id = args.id
        type = args.type
        if name is not None:
            name = name.split(",")
        if key is not None:
            key = key.split(",")
        if phase is not None:
            phase = phase.split(",")
        if id is not None:
            id = id.split(",")
        if type is not None:
            type = type.split(",")
        wf = Workflow(id=wf_id)
        if key is not None:
            steps = wf.query_step_by_key(key, name, phase, id, type)
        else:
            steps = wf.query_step(name, key, phase, id, type)
        for step in steps:
            if step.type in ["StepGroup"]:
                continue
            key = step.key if step.key is not None else ""
            dur = format_time_delta(step.get_duration())
            t = []
            t.append(["Step:", step.displayName])
            t.append(["ID:", step.id])
            t.append(["Key:", key])
            t.append(["Type:", step.type])
            t.append(["Phase:", step.phase])
            format_print_table(t)
            if hasattr(step, "outputs"):
                if hasattr(step.outputs, "parameters"):
                    print("Output parameters:")
                    for name, par in step.outputs.parameters.items():
                        if name[:6] == "dflow_":
                            continue
                        print("%s: %s" % (name, par.value))
                    print()
                if hasattr(step.outputs, "artifacts"):
                    print("Output artifacts:")
                    for name, art in step.outputs.artifacts.items():
                        if name[:6] == "dflow_" or name == "main-logs":
                            continue
                        key = ""
                        if hasattr(art, "s3"):
                            key = art.s3.key
                        elif hasattr(art, "oss"):
                            key = art.oss.key
                        print("%s: %s" % (name, key))
                    print()
            print()
    elif args.command == "getkeys":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        keys = wf.query_keys_of_steps()
        print("\n".join(keys))
    elif args.command == "delete":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.delete()
    elif args.command == "resubmit":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.resubmit()
    elif args.command == "resume":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.resume()
    elif args.command == "retry":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        if args.step is not None:
            step = wf.query_step(id=args.step)[0]
            step.retry()
        else:
            wf.retry()
    elif args.command == "stop":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.stop()
    elif args.command == "suspend":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.suspend()
    elif args.command == "terminate":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.terminate()
    elif args.command == "wait":
        wf_id = args.ID
        wf = Workflow(id=wf_id)
        wf.wait(interval=args.interval)
    elif args.command == "download":
        assert args.key is not None or args.urn is not None, \
            "one of -k/--key and -u/--urn must be specified"
        art = S3Artifact(key=args.key, urn=args.urn)
        path = download_artifact(art, path=args.path)
        print("Downloaded artifact to %s" % path)
    elif args.command == "upload":
        if "=" in args.path:
            path = {}
            for p in args.path.split(","):
                k, v = p.split("=")
                path[k] = v
        else:
            path = args.path.split(",")
        tags = None
        if args.tag is not None:
            tags = args.tag.split(",")
        properties = None
        if args.property is not None:
            properties = {}
            for p in args.property.split(","):
                k, v = p.split("=")
                properties[k] = v
        art = upload_artifact(
            path, namespace=args.namespace, dataset_name=args.name,
            description=args.description, tags=tags, properties=properties)
        print("Storage key: %s" % art.key)
        if art.urn:
            print("Dataset URN: %s" % art.urn)
    elif args.command == "submit":
        with open(args.FILE, "r") as f:
            wf = Workflow.from_yaml(f.read())
        if args.name is not None:
            wf.id = args.name
        if args.detach:
            config["detach"] = True
        wf.submit()
    elif args.command == "create":
        if args.resource == "secret":
            s = Secret(args.value, args.name, args.key)
            print("Secret (name: %s, key: %s) created" % (s.secret_name,
                                                          s.secret_key))


if __name__ == "__main__":
    main()



import time

from dflow import (OutputArtifact, OutputParameter, Outputs, Step, Steps,
                   Workflow, if_expression)
from dflow.python import (OP, OPIO, Artifact, OPIOSign, PythonOPTemplate,
                          upload_packages)

if "__file__" in locals():
    upload_packages.append(__file__)


class Random(OP):
    @classmethod
    def get_input_sign(cls):
        return OPIOSign()

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            "is_head": bool,
            "msg1": str,
            "msg2": str,
            "foo": Artifact(str),
            "bar": Artifact(str)
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        open("foo.txt", "w").write("head")
        open("bar.txt", "w").write("tail")
        is_head = False
        return OPIO({
            "is_head": is_head,
            "msg1": "head",
            "msg2": "tail",
            "foo": "foo.txt",
            "bar": "bar.txt"
        })


def test_conditional_outputs():
    steps = Steps("steps", outputs=Outputs(
        parameters={"msg": OutputParameter()},
        artifacts={"res": OutputArtifact()}))

    random_step = Step(
        name="random",
        template=PythonOPTemplate(Random, image="python:3.8")
    )
    steps.add(random_step)

    steps.outputs.parameters["msg"].value_from_expression = if_expression(
        _if=random_step.outputs.parameters["is_head"],
        _then=random_step.outputs.parameters["msg1"],
        _else=random_step.outputs.parameters["msg2"])

    steps.outputs.artifacts["res"].from_expression = if_expression(
        _if=random_step.outputs.parameters["is_head"],
        _then=random_step.outputs.artifacts["foo"],
        _else=random_step.outputs.artifacts["bar"])

    wf = Workflow(name="conditional")
    step = Step("conditional", template=steps)
    wf.add(step)
    wf.submit()

    while wf.query_status() in ["Pending", "Running"]:
        time.sleep(1)

    assert(wf.query_status() == "Succeeded")
    step = wf.query_step(name="conditional")[0]
    assert(step.phase == "Succeeded")
    assert step.outputs.parameters["msg"].value == "tail"


if __name__ == "__main__":
    test_conditional_outputs()



import time
from typing import List

from dflow import Step, Workflow, argo_len, argo_range
from dflow.python import (OP, OPIO, Artifact, OPIOSign, PythonOPTemplate,
                          Slices, upload_packages)

if "__file__" in locals():
    upload_packages.append(__file__)


class Prepare(OP):
    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign()

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'data': List[int]
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        op_out = OPIO({
            'data': list(range(1, 18))
        })
        return op_out


class Process(OP):
    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'data': int
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            "output": Artifact(str)
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        with open("%s.txt" % op_in["data"], "w") as f:
            f.write("Hello")
        return OPIO({
            "output": "%s.txt" % op_in["data"]
        })


def test_group_size():
    wf = Workflow("slices")
    prepare = Step("prepare",
                   PythonOPTemplate(Prepare, image="python:3.8"))
    wf.add(prepare)
    process = Step(
        "process",
        PythonOPTemplate(
            Process,
            image="python:3.8",
            slices=Slices(
                "{{item}}",
                input_parameter=["data"],
                output_artifact=["output"],
                pool_size=1,
                group_size=10,
            ),
        ),
        parameters={"data": prepare.outputs.parameters["data"]},
        with_param=argo_range(argo_len(prepare.outputs.parameters["data"])),
    )
    wf.add(process)
    wf.submit()

    while wf.query_status() in ["Pending", "Running"]:
        time.sleep(1)

    assert(wf.query_status() == "Succeeded")


if __name__ == "__main__":
    test_group_size()



import json
from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

import jsonpickle

from ..common import S3Artifact
from ..config import config
from ..io import PVC


class nested_dict:
    def __init__(self, type):
        self.type = type

    def __eq__(self, other):
        if not isinstance(other, nested_dict):
            return False
        return self.type == other.type


NestedDict = {
    str: nested_dict(str),
    Path: nested_dict(Path),
}

ArtifactAllowedTypes = [str, Path, Set[str], Set[Path], List[str], List[Path],
                        Dict[str, str], Dict[str, Path], NestedDict[str],
                        NestedDict[Path]]


def type_to_str(type):
    if hasattr(type, "__module__") and hasattr(type, "__name__"):
        if type.__module__ == "builtins":
            return type.__name__
        else:
            return "%s.%s" % (type.__module__, type.__name__)
    else:
        return str(type)


class Artifact:
    """
    OPIO signature of artifact

    Args:
        type: str, Path, Set[str], Set[Path], List[str], List[Path],
            Dict[str, str], Dict[str, Path], NestedDict[str] or
            NestedDict[Path]
        archive: compress format of the artifact, None for no compression
        save: place to store the output artifact instead of default storage,
            can be a list
        optional: optional input artifact or not
        global_name: global name of the artifact within the workflow
    """

    def __init__(
            self,
            type: Any,
            archive: str = "default",
            save: List[Union[PVC, S3Artifact]] = None,
            optional: bool = False,
            global_name: Optional[str] = None,
            sub_path: bool = True,
            **kwargs,
    ) -> None:
        self.type = type
        if archive == "default":
            archive = config["archive_mode"]
        self.archive = archive
        self.save = save
        self.optional = optional
        self.global_name = global_name
        self.sub_path = sub_path
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __setattr__(self, key, value):
        if key == "type":
            assert (value in ArtifactAllowedTypes), "%s is not allowed" \
                                                    "artifact type, only %s " \
                                                    "are allowed." % (
                                                        value,
                                                        ArtifactAllowedTypes)
        super().__setattr__(key, value)

    def to_str(self):
        return "Artifact(type=%s, optional=%s, sub_path=%s)" % (
            type_to_str(self.type), self.optional, self.sub_path)

    def __call__(self, *args, **kwargs):
        self.type(*args, **kwargs)


class Parameter:
    """
    OPIO signature of parameter

    Args:
        type: parameter type
        global_name: global name of the parameter within the workflow
        default: default value of the parameter
    """

    def __init__(
            self,
            type: Any,
            global_name: Optional[str] = None,
            **kwargs,
    ) -> None:
        self.type = type
        self.global_name = global_name
        if "default" in kwargs:
            self.default = kwargs["default"]
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def to_str(self):
        default = ""
        if hasattr(self, "default"):
            try:
                default = ", default=%s" % json.dumps(self.default)
            except Exception:
                default = ", default=jsonpickle.loads('%s')" % \
                    jsonpickle.dumps(self.default)
        return "Parameter(type=%s%s)" % (type_to_str(self.type), default)


class BigParameter:
    """
    OPIO signature of big parameter

    Args:
        type: parameter type
    """

    def __init__(
            self,
            type: Any,
            **kwargs,
    ) -> None:
        self.type = type
        if "default" in kwargs:
            self.default = kwargs["default"]
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def to_str(self):
        default = ""
        if hasattr(self, "default"):
            try:
                default = ", default=%s" % json.dumps(self.default)
            except Exception:
                default = ", default=jsonpickle.loads('%s')" % \
                    jsonpickle.dumps(self.default)
        return "BigParameter(type=%s%s)" % (type_to_str(self.type), default)


class OPIOSign(MutableMapping):
    """The signature of OPIO.
    A signature of OPIO includes the key and its typing
    """

    def __init__(
            self,
            *args,
            **kwargs
    ):
        self._data = {}
        self._data = dict(*args, **kwargs)

    def __getitem__(
            self,
            key: str,
    ) -> Any:
        """Get the type hint of the key
        """
        return self._data[key]

    def __setitem__(
            self,
            key: str,
            value: Any,
    ) -> None:
        """Set the type hint of the key
        """
        self._data[key] = value

    def __delitem__(
            self,
            key: str,
    ) -> None:
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return str(self._data)


class OPIO(MutableMapping):
    def __init__(
            self,
            *args,
            **kwargs
    ):
        self._data = {}
        self._data = dict(*args, **kwargs)

    def __getitem__(
            self,
            key: str,
    ) -> Any:
        return self._data[key]

    def __setitem__(
            self,
            key: str,
            value: Any,
    ) -> None:
        self._data[key] = value

    def __delitem__(
            self,
            key: str,
    ) -> None:
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return str(self._data)



from .v1alpha1_artifact import V1alpha1Artifact
from .v1alpha1_parameter import V1alpha1Parameter
from .v1alpha1_retry_strategy import V1alpha1RetryStrategy
from .v1alpha1_sequence import V1alpha1Sequence
from .v1alpha1_value_from import V1alpha1ValueFrom

__all__ = ["V1alpha1Artifact", "V1alpha1Parameter",
           "V1alpha1RetryStrategy", "V1alpha1Sequence", "V1alpha1ValueFrom"]



# coding: utf-8

"""
    Argo Server API

    You can get examples of requests and responses by using the CLI with `--gloglevel=9`, e.g. `argo list --gloglevel=9`  # noqa: E501

    The version of the OpenAPI document: v2.12.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from argo.workflows.client.configuration import Configuration


class V1alpha1ValueFrom(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'default': 'str',
        'event': 'str',
        'jq_filter': 'str',
        'json_path': 'str',
        'parameter': 'str',
        'path': 'str',
        'supplied': 'object',
        'expression': 'str'
    }

    attribute_map = {
        'default': 'default',
        'event': 'event',
        'jq_filter': 'jqFilter',
        'json_path': 'jsonPath',
        'parameter': 'parameter',
        'path': 'path',
        'supplied': 'supplied',
        'expression': 'expression'
    }

    def __init__(self, default=None, event=None, jq_filter=None, json_path=None, parameter=None, path=None, supplied=None, expression=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1ValueFrom - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._default = None
        self._event = None
        self._jq_filter = None
        self._json_path = None
        self._parameter = None
        self._path = None
        self._supplied = None
        self._expression = None
        self.discriminator = None

        if default is not None:
            self.default = default
        if event is not None:
            self.event = event
        if jq_filter is not None:
            self.jq_filter = jq_filter
        if json_path is not None:
            self.json_path = json_path
        if parameter is not None:
            self.parameter = parameter
        if path is not None:
            self.path = path
        if supplied is not None:
            self.supplied = supplied
        if expression is not None:
            self.expression = expression

    @property
    def default(self):
        """Gets the default of this V1alpha1ValueFrom.  # noqa: E501

        Default specifies a value to be used if retrieving the value from the specified source fails  # noqa: E501

        :return: The default of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: str
        """
        return self._default

    @default.setter
    def default(self, default):
        """Sets the default of this V1alpha1ValueFrom.

        Default specifies a value to be used if retrieving the value from the specified source fails  # noqa: E501

        :param default: The default of this V1alpha1ValueFrom.  # noqa: E501
        :type: str
        """

        self._default = default

    @property
    def event(self):
        """Gets the event of this V1alpha1ValueFrom.  # noqa: E501

        Selector (https://github.com/antonmedv/expr) that is evaluated against the event to get the value of the parameter. E.g. `payload.message`  # noqa: E501

        :return: The event of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: str
        """
        return self._event

    @event.setter
    def event(self, event):
        """Sets the event of this V1alpha1ValueFrom.

        Selector (https://github.com/antonmedv/expr) that is evaluated against the event to get the value of the parameter. E.g. `payload.message`  # noqa: E501

        :param event: The event of this V1alpha1ValueFrom.  # noqa: E501
        :type: str
        """

        self._event = event

    @property
    def jq_filter(self):
        """Gets the jq_filter of this V1alpha1ValueFrom.  # noqa: E501

        JQFilter expression against the resource object in resource templates  # noqa: E501

        :return: The jq_filter of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: str
        """
        return self._jq_filter

    @jq_filter.setter
    def jq_filter(self, jq_filter):
        """Sets the jq_filter of this V1alpha1ValueFrom.

        JQFilter expression against the resource object in resource templates  # noqa: E501

        :param jq_filter: The jq_filter of this V1alpha1ValueFrom.  # noqa: E501
        :type: str
        """

        self._jq_filter = jq_filter

    @property
    def json_path(self):
        """Gets the json_path of this V1alpha1ValueFrom.  # noqa: E501

        JSONPath of a resource to retrieve an output parameter value from in resource templates  # noqa: E501

        :return: The json_path of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: str
        """
        return self._json_path

    @json_path.setter
    def json_path(self, json_path):
        """Sets the json_path of this V1alpha1ValueFrom.

        JSONPath of a resource to retrieve an output parameter value from in resource templates  # noqa: E501

        :param json_path: The json_path of this V1alpha1ValueFrom.  # noqa: E501
        :type: str
        """

        self._json_path = json_path

    @property
    def parameter(self):
        """Gets the parameter of this V1alpha1ValueFrom.  # noqa: E501

        Parameter reference to a step or dag task in which to retrieve an output parameter value from (e.g. '{{steps.mystep.outputs.myparam}}')  # noqa: E501

        :return: The parameter of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: str
        """
        return self._parameter

    @parameter.setter
    def parameter(self, parameter):
        """Sets the parameter of this V1alpha1ValueFrom.

        Parameter reference to a step or dag task in which to retrieve an output parameter value from (e.g. '{{steps.mystep.outputs.myparam}}')  # noqa: E501

        :param parameter: The parameter of this V1alpha1ValueFrom.  # noqa: E501
        :type: str
        """

        self._parameter = parameter

    @property
    def path(self):
        """Gets the path of this V1alpha1ValueFrom.  # noqa: E501

        Path in the container to retrieve an output parameter value from in container templates  # noqa: E501

        :return: The path of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path of this V1alpha1ValueFrom.

        Path in the container to retrieve an output parameter value from in container templates  # noqa: E501

        :param path: The path of this V1alpha1ValueFrom.  # noqa: E501
        :type: str
        """

        self._path = path

    @property
    def supplied(self):
        """Gets the supplied of this V1alpha1ValueFrom.  # noqa: E501

        Supplied value to be filled in directly, either through the CLI, API, etc.  # noqa: E501

        :return: The supplied of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: object
        """
        return self._supplied

    @supplied.setter
    def supplied(self, supplied):
        """Sets the supplied of this V1alpha1ValueFrom.

        Supplied value to be filled in directly, either through the CLI, API, etc.  # noqa: E501

        :param supplied: The supplied of this V1alpha1ValueFrom.  # noqa: E501
        :type: object
        """

        self._supplied = supplied

    @property
    def expression(self):
        """Gets the expression of this V1alpha1ValueFrom.  # noqa: E501

        Supplied value to be filled in directly, either through the CLI, API, etc.  # noqa: E501

        :return: The expression of this V1alpha1ValueFrom.  # noqa: E501
        :rtype: object
        """
        return self._expression

    @expression.setter
    def expression(self, expression):
        """Sets the expression of this V1alpha1ValueFrom.

        Supplied value to be filled in directly, either through the CLI, API, etc.  # noqa: E501

        :param expression: The expression of this V1alpha1ValueFrom.  # noqa: E501
        :type: object
        """

        self._expression = expression

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1ValueFrom):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1ValueFrom):
            return True

        return self.to_dict() != other.to_dict()
